import sys
import os
import importlib
import __main__

class Pigeon(object):
    def __init__(self, *args, **kwargs):
        pass
    
    
    @classmethod
    def get_temp_filename(cls):
        """the name of the temporary file wing will use in write_temp_file()
        
        Sub-classes can override this is they want a unique name for their
        Pigeon.
        """
        return('wing_output_text.txt')
    
    

    @classmethod
    def get_temp_filepath(cls):
        """Returns the full path of the temp file on the local system."""
        return os.path.join(os.environ['TMP'], cls.get_temp_filename())



    @classmethod
    def write_temp_file(cls, txt):
        """writes the input text to the get_temp_filepath()"""
    
        # Save the text to a temp file. If we're dealing with mel, make sure it
        # ends with a semicolon, or Maya could become angered!
        #txt = get_wing_text()
        temp_path = cls.get_temp_filepath()
        f = open(temp_path, "wb")
    
        print('writing temp file:{}'.format(temp_path))
        f.write(txt.encode())
    
        f.close()

        temp_path = temp_path.replace("\\", "/")
        return temp_path
    
    

    @staticmethod
    def read_file(file_path):
        """Executes the python code stored in the file path.
        
        Args:
            file_path (string) : The absolute file path of the py file to read
        
        """
        
        print("WING: executing code from file {}\n".format(file_path))
        if os.access(file_path, os.F_OK):
            # execute the file contents in Maya:
            with open(file_path, "rb") as f:
                data = f.read()
                data = data.decode()
                exec(data, __main__.__dict__, __main__.__dict__) 

        else:
            print("No Wing-generated temp file exists: " + file_path)
            
            
            
    @classmethod
    def post_module_import(cls, module):
        """called if a Pigeon.import_module() is successful.
        
        The default behavior is to call the run() on the module if one
        exists. Override this method in a sub-class if you have custom logic
        for how you want what happens after a module has been
        imported/reloaded.
        
        Args:
            module : The module that was imported/reloaded from import_module()
        """
        if hasattr(module, 'run'):
            module.run()
            
            

    @classmethod
    def import_module(cls, module_name, file_path):
        """Attempts to import/reload the input module name and execute any run()
        
        If importing/reloading fails, then the file path is read (if
        defined). If sub-classes don't want to call run() they should
        override post_module_import().
        
        Args:
            module_name (string) : The name of the module relative to any
            package namespace.
            file_path (string) : The absolute file path to the py file.
        """
        
        print('\n')
        imported = module_name in sys.modules
        if imported:
            print('reloading module:{0}'.format(module_name))
            importlib.reload(sys.modules[module_name])
        else:
            try:
                print('Attempting module import of:{0}'.format(module_name))
                importlib.import_module(module_name)
            except ModuleNotFoundError:
                if file_path:
                    cls.read_file(file_path)


        if module_name in sys.modules:
            cls.post_module_import(sys.modules[module_name])
    
    

    def can_dispatch(self):
        """Check if conditions are right to send code to application
        
        can_dispatch() is used to determine what dispatcher wing will use
        with when there's no active dispatcher found.
        """
        raise NotImplementedError
    
    
    def owns_process(self, process):
        """Returns true if the process is the pigeon's target application
        
        This is used when an external application is connects to wing. When True
        is returned the Dispatcher becomes the active dispatcher to send commands
        to.
        
        Args:
            process (psutils.Process) : The node to remove the data from.  
        """
        raise NotImplementedError
    

    def send(self, highlighted_text, module_path, file_path, doc_type):
        """The main entry point for sending content from wing to an external app
        
        sub-classes should override this with application specfic logic for how
        the data is sent to an external application.        
        """
        raise NotImplementedError
    
    
    def send_python_command(self, command_string):
        """Send a custom python command to the target application
        
        Returns:
            bool : True if the command was successfully sent.        
        """
        raise NotImplementedError





#########################################################################
""" wingdbstub.py -- Start debugging Python programs from outside of Wing

Copyright (c) 1999-2018, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel and John P. Ehresman

Usage:
-----

This file is used to initiate debug in Python code without launching
that code from Wing.  To use it, edit the configuration values below,
start Wing and configure it to listen for outside debug connections,
and then add the following line to your code:

  import wingdbstub

Debugging will start immediately after this import statement is reached.

For details, see Debugging Externally Launched Code in the manual.

"""
#########################################################################

import sys
orig_sys_modules = list(sys.modules.keys())
orig_sys_path = list(sys.path)
orig_meta_path = list(sys.meta_path)

import os
if sys.version_info >= (3, 7):
  import importlib
else:
  import imp

#------------------------------------------------------------------------
# Required configuration values (some installers set this automatically)

# This should be the full path to your Wing installation.  On macOS, use 
# the full path of the Wing application bundle, for example 
# /Applications/WingPro.app.  When set to None, the environment variable 
# WINGHOME is used instead.  
WINGHOME = None
for path in [r"C:\Program Files\Wing Pro 9",
                 r"C:\Program Files\Wing Pro 8",
                 r"C:\Program Files\Wing Pro 7"
                 ]:
  if os.path.exists(path):
    WINGHOME = path
    break

#------------------------------------------------------------------------
# Optional configuration values:  The named environment variables, if set, 
# will override these settings.

# Set this to 1 to disable all debugging; 0 to enable debugging
# (WINGDB_DISABLED environment variable)
kWingDebugDisabled = 0

# Host:port of the IDE within which to debug: As configured in the IDE
# with the Server Port preference
# (WINGDB_HOSTPORT environment variable)
kWingHostPort = 'localhost:50005'

# Port on which to listen for connection requests, so that the
# IDE can attach or reattach to the debug process after it has started.
# Set this to '-1' to disable listening for connection requests.
# This is only used when the debug process is not attached to
# an IDE or the IDE has dropped its connection. The configured
# port can optionally be added to the IDE's Common Attach Hosts
# preference. Note that a random port is used instead if the given 
# port is already in use.
# (WINGDB_ATTACHPORT environment variable)
kAttachPort = '50015'

# Set this to a filename to log verbose information about the debugger
# internals to a file.  If the file does not exist, it will be created
# as long as its enclosing directory exists and is writeable.  Use 
# "<stderr>" or "<stdout>" to write to stderr or stdout.  Note that 
# "<stderr>" may cause problems on Windows if the debug process is not 
# running in a console.
# (WINGDB_LOGFILE environment variable)
kLogFile = None

# Produce a tremendous amount of logging from the debugger internals.
# Do not set this unless instructed to do so by Wingware support.  It
# will slow the debugger to a crawl.
# (WINGDB_LOGVERYVERBOSE environment variable)
kLogVeryVerbose = 0

# Set this to 1 when debugging embedded scripts in an environment that
# creates and reuses a Python instance across multiple script invocations.  
# It turns off automatic detection of program quit so that the debug session
# can span multiple script invocations.  When this is turned on, you may
# need to call ProgramQuit() on the debugger object to shut down the
# debugger cleanly when your application exits or discards the Python
# instance.  If multiple Python instances are created in the same run,
# only the first one will be able to debug unless it terminates debug
# and the environment variable WINGDB_ACTIVE is unset before importing
# this module in the second or later Python instance.  See the Wing
# manual for details.
# (WINGDB_EMBEDDED environment variable)
kEmbedded = 1

# Security Token configuration

# Wing uses a security token to control access from the debug process.
# In cases where the debugger is running as the same user and on the same
# host as the IDE, no further configuration is needed because the token
# stored in wingdebugpw in the Settings Directory shown in Wing's About
# box is used by both ends of the connection.  In other cases, Wing will
# wrote a new security token to the Settings Directory for the user running
# the debug process.  An unknown security token will be rejected, but Wing
# will offer to accept future connections using an that token.  To avoid
# this, copy the file wingdebugpw from the Settings Directory shown in Wing's
# About box to the same directory as this copy of wingdbstub.py, or set
# kSecurityToken below to the contents of that file.  This is also needed
# when Wing cannot write to the Settings Directory, or the contents of that
# directory are not persisted between debug runs, as is the case with some
# types of containers.

# The security token to use, overriding any other value.  If None, the token is 
# instead read from the file configured with kFWFilePath and kPWFileName below.  
# Disk permissions of any copy of wingdbstub.py with kSecurityToken set should 
# be adjusted to avoid unauthorized access by other users.
# (WINGDB_SECURITYTOKEN environment variable)
kSecurityToken = None

# Path to search for the debugger security token and the name of the file
# to use.  These values are for advanced use cases and rarely need to be changed.  
# In kPWFilePath, the value '$<winguserprofile>' is replaced by the User Settings 
# Directory for the user that is running the debug process.
# (WINGDB_PWFILEPATH and WINGDB_PWFILENAME environment variables)
kPWFilePath = [os.path.dirname(__file__), '$<winguserprofile>']
kPWFileName = 'wingdebugpw'

# Whether to exit when the debugger fails to run or to connect with the IDE
# By default, execution continues without debug or without connecting to
# the IDE.
# (WINGDB_EXITONFAILURE environment variable)
kExitOnFailure = 0

#------------------------------------------------------------------------
# Find Wing debugger installation location

# Check environment:  Must have WINGHOME defined if still == None
if WINGHOME == None:
  if 'WINGHOME' in os.environ:
    WINGHOME=os.environ['WINGHOME']
  else:
    msg = '\n'.join(["*******************************************************************", 
                     "Error: Could not find Wing installation!  You must set WINGHOME or edit", 
                     "wingdbstub.py where indicated to point it to the location where", 
                     "Wing is installed.\n"])
    sys.stderr.write(msg)
    raise ImportError(msg)
  
WINGHOME = os.path.expanduser(WINGHOME)

def NP_FindActualWingHome(winghome):
  """ Find the actual directory to use for winghome.  Needed on macOS
  where the .app directory is the preferred dir to use for WINGHOME and
  .app/Contents/MacOS is accepted for backward compatibility. """
  
  if sys.platform != 'darwin':
    return winghome
  
  app_dir = None
  if os.path.isdir(winghome):
    if winghome.endswith('/'):
      wo_slash = winghome[:-1]
    else:
      wo_slash = winghome
      
    if wo_slash.endswith('.app'):
      app_dir = wo_slash
    elif wo_slash.endswith('.app/Contents/MacOS'):
      app_dir = wo_slash[:-len('/Contents/MacOS')]
    
  if app_dir and os.path.isdir(os.path.join(app_dir, 'Contents', 'Resources')):
    return os.path.join(app_dir, 'Contents', 'Resources')
  
  return winghome
  
WINGHOME = NP_FindActualWingHome(WINGHOME)
os.environ['WINGHOME'] = WINGHOME

#-----------------------------------------------------------------------
def NP_LoadModuleFromBootstrap(winghome, modname):
  """Load a module from the installation bootstrap directory.  Assumes that
  imports don't change sys.path.  The modules are unloaded from sys.modules
  so they can be loaded again later from an update."""
  
  # Limited to simple module loads
  assert '.' not in modname
  
  orig_sys_path = sys.path[:]
  orig_modules = set(sys.modules)
  
  dirname = winghome + '/bootstrap'
  sys.path.insert(0, dirname)
  
  code = 'import %s' % modname
  exec(code)
  
  new_modules = set(sys.modules)
  new_modules.difference_update(orig_modules)
  for mod in new_modules:
    del sys.modules[mod]
    
  sys.path = orig_sys_path
  
  return locals()[modname]

#------------------------------------------------------------------------
# Set debugger if it hasn't been set -- this is to handle module reloading
# In the reload case, the debugger variable will be set to something
try:
  debugger
except NameError:
  debugger = None
  
#-----------------------------------------------------------------------
# Unset WINGDB_ACTIVE env if it was inherited from another process
# XXX Would be better to be able to call getpid() on dbgtracer but can't access it yet
if 'WINGDB_ACTIVE' in os.environ and os.environ['WINGDB_ACTIVE'] != str(os.getpid()):
  del os.environ['WINGDB_ACTIVE']

#-----------------------------------------------------------------------
# Start debugging if not disabled and this module has never been imported
# before
if (not kWingDebugDisabled and debugger is None and
    'WINGDB_DISABLED' not in os.environ and 
    'WINGDB_ACTIVE' not in os.environ):

  exit_on_fail = 0
  
  try:
    # Obtain exit if fails value
    exit_on_fail = os.environ.get('WINGDB_EXITONFAILURE', kExitOnFailure)
    
    # Obtain configuration for log file to use, if any
    logfile = os.environ.get('WINGDB_LOGFILE', kLogFile)
    if logfile == '-' or logfile == None or len(logfile.strip()) == 0:
      logfile = None

    very_verbose_log = os.environ.get('WINGDB_LOGVERYVERBOSE', kLogVeryVerbose)
    if type(very_verbose_log) == type('') and very_verbose_log.strip() == '':
      very_verbose_log = 0
      
    # Determine remote host/port where the IDE is running
    hostport = os.environ.get('WINGDB_HOSTPORT', kWingHostPort)
    colonpos = hostport.find(':')
    host = hostport[:colonpos]
    port = int(hostport[colonpos+1:])
  
    # Determine port to listen on locally for attach requests
    attachport = int(os.environ.get('WINGDB_ATTACHPORT', kAttachPort))
  
    # Check if running embedded script
    embedded = int(os.environ.get('WINGDB_EMBEDDED', kEmbedded))
  
    # Obtain security token
    security_token = os.environ.get('WINGDB_SECURITYTOKEN', kSecurityToken)
    
    # Obtain debug password file search path
    if 'WINGDB_PWFILEPATH' in os.environ:
      pwfile_path = os.environ['WINGDB_PWFILEPATH'].split(os.pathsep)
    else:
      pwfile_path = kPWFilePath
    
    # Obtain debug password file name
    if 'WINGDB_PWFILENAME' in os.environ:
      pwfile_name = os.environ['WINGDB_PWFILENAME']
    else:
      pwfile_name = kPWFileName
    
    # Set up temporary log for errors from merge importer Setup
    class CTmpLog:
      def __init__(self):
        self.fErrors = []
      def write(self, s):
        self.fErrors.append(s)
    tmp_log = CTmpLog()
    
    # Set up the meta importer; everything after this point can be updated
    bootstrap_utils = NP_LoadModuleFromBootstrap(WINGHOME, 'bootstrap_utils')
    winghome, user_settings = bootstrap_utils.NP_SetupWingHomeModule(WINGHOME)
    meta = bootstrap_utils.NP_CreateMetaImporter(WINGHOME, user_settings, 'dbg',
                                                 tmp_log)
    import _winghome
    _winghome.kWingHome = winghome
    _winghome.kUserSettings = user_settings
    
    # Find the netserver module and create an error stream
    from debug.tserver import startdebug
    netserver = startdebug.FindNetServerModule(WINGHOME, user_settings)
    err = startdebug.CreateErrStream(netserver, logfile, very_verbose_log)
    
    # Write any temporary log entries
    for s in tmp_log.fErrors:
      err.write(s)
      
    # Create debug server
    debugger = netserver.CNetworkServer(host, port, attachport, err,
                                        security_token=security_token, 
                                        pwfile_path=pwfile_path,
                                        pwfile_name=pwfile_name,
                                        autoquit=not embedded)
    
    # Restore module and path environment
    from debug.tserver import startdebug
    startdebug.RestoreEnvironment(orig_sys_modules, orig_sys_path, orig_meta_path)
    
    # Start debugging
    debugger.StartDebug(stophere=0)
    os.environ['WINGDB_ACTIVE'] = str(os.getpid())
    if debugger.ChannelClosed():
      raise ValueError('Not connected')
    
  except:
    if exit_on_fail:
      raise
    else:
      pass

#-----------------------------------------------------------------------
def Ensure(require_connection=1, require_debugger=1):
  """ Ensure the debugger is started and attempt to connect to the IDE if
  not already connected.  Will raise a ValueError if:
  
  * the require_connection arg is true and the debugger is unable to connect
  * the require_debugger arg is true and the debugger cannot be loaded
  
  If SuspendDebug() has been called through the low-level API, calling
  Ensure() resets the suspend count to zero and additional calls to
  ResumeDebug() will be ignored until SuspendDebug() is called again.
  
  Note that a change to the host & port to connect to will only
  be use if a new connection is made.
  
  """
  
  if debugger is None:
    if require_debugger:
      raise ValueError("No debugger")
    return

  hostport = os.environ.get('WINGDB_HOSTPORT', kWingHostPort)
  colonpos = hostport.find(':')
  host = hostport[:colonpos]
  port = int(hostport[colonpos+1:])
  
  resumed = debugger.ResumeDebug()
  while resumed > 0:
    resumed = debugger.ResumeDebug()
  
  debugger.SetClientAddress((host, port))  
  
  if not debugger.DebugActive():
    debugger.StartDebug()
  elif debugger.ChannelClosed():
    debugger.ConnectToClient()
    
  if require_connection and debugger.ChannelClosed():
    raise ValueError('Not connected')
    


from .pigeon import *
import socket


#import below are used maya side to receive commands
import sys
import __main__
import importlib
_MAYA_ACTIVE = False
try:
    import maya.OpenMaya as om
    _MAYA_ACTIVE = False
except:
    pass



class MayaPigeon(Pigeon):
    commandPort = 6000

    
    def __init__(self, *args, **kwargs):
        super(MayaPigeon, self).__init__(*args, **kwargs)

        
    @classmethod
    def get_temp_filename(cls):
        return('wing_maya_temp.txt')


    def get_socket(self):
        
        # The commandPort you opened in userSetup.py Make sure this matches!
        m_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    
        # Now ping Maya over the command-port
        try:
            # Make our socket-> Maya connection: There are different connection ways
            # which vary between machines, so sometimes you need to try different
            # solutions to get it to work... :-S
            #mSocket.connect(("127.0.0.1", commandPort))
            #mSocket.connect(("::1",commandPort)) #works!
            m_socket.connect(("127.0.0.1", MayaPigeon.commandPort))

        except Exception as e:
            print('Connection to Maya failed: {}'.format(e))
            m_socket.close()
            m_socket = None

        finally:
            return m_socket


    def can_dispatch(self):
        """Check if conditions are right to send code to application
        
        can_dispatch() is used to determine what dispatcher wing will use
        with when there's no active dispatcher found.
        """
        
        m_socket = self.get_socket()
        if m_socket is not None:
            m_socket.close()
            return True
        else:
            return False
        

    def owns_process(self, process):
        """Returns true if the process is the pigeons target application
        
        This is used when an external application is connects to wing. When True
        is returned the Dispatcher becomes the active dispatcher to send commands
        to.
        
        Args:
            process (psutils.Process) : The node to remove the data from.  
        """
        return 'maya' in process.name()
      

    @classmethod
    def read_file(cls, file_path, doc_type=''):
        """
        Evaluate the temp file on disk, made by Wing, in Maya.
    
        Args:
            file_path (str) : The absolute path to the file to load
            doc_type (str) : Supports either 'python' or 'mel'
        """
        
        print_lines = False
        if not doc_type:
            doc_type = 'python'
            
        if doc_type == 'python':
            super(MayaPigeon, cls).read_file(file_path)
            #Pigeon.read_file(file_path)
        else:
            print("MayaPigeon : Running MEL code from file {}\n".format(file_path))
            if os.access(file_path, os.F_OK):
                if print_lines:
                    #temp data is likely highlighted code from Wing
                    #so let's print it for review.
                    with open(file_path, "rb") as f:
                        for line in f.readlines():
                            print(line.rstrip())
                    print("\n"),

                melCmd = 'source "%s"' % file_path
                # This causes the "// Result: " line to show up in the Script Editor:
                om.MGlobal.executeCommand(melCmd, True, True)
            else:
                print("No Wing-generated temp file exists: " + file_path)
            

    @classmethod
    def receive(cls, module_path, doc_type, file_path):
        print("{} {} {}".format(module_path, doc_type, file_path))
        if not module_path:
            cls.read_file(file_path, doc_type=doc_type)

        elif 'python' in doc_type:
            cls.import_module(module_path, file_path)


    def send(self, highlighted_text, module_path, file_path, doc_type):
        """The main entry point for sending content from wing to an external app"""
        m_socket = self.get_socket()
        if m_socket is None:
            print("Can't communicate with Maya!")
            return
        
        if 'python' not in doc_type and file_path.endswith('mel'):
            doc_type = 'mel'
            module_path = ''        
        
        if highlighted_text:
            module_path = ''
            if doc_type == 'mel' and not highlighted_text.endswith(';'):
                highlighted_text += ';'
            
            file_path = self.write_temp_file(highlighted_text)

        try:
            command = u"import wingcarrier.pigeons; wingcarrier.pigeons.MayaPigeon.receive(\'{}\',\'{}\',\'{}\')".format(module_path, doc_type, file_path)
            print(command)
            m_socket.send(command.encode())
        except Exception as e:
            print("Maya socket errored:{}".format(e))
            
        finally:
            m_socket.close()
            
            
    def send_python_command(self, command_string):
        m_socket = self.get_socket()
        if m_socket is None:
            print("Can't connect to Maya!")
            return False
        
        success = False
        try:
            #command = u'python("{}")'.format(command_string)
            m_socket.send(command_string.encode())
            success = True
        except Exception as e:
            print("Maya socket errored:{}".format(e))
        finally:
            m_socket.close()
              
        return success




import inspect
import subprocess

PSUTIL_EXISTS = False
try:
    
    import psutil
    PSUTIL_EXISTS = True
except:
    #this will fail when using the module in cascadeur
    print("Missing python package 'psutil'. CascadeurPigeon functionality limited to receiving")
    pass



CSC_EXISTS = False
try:
    import csc
    CSC_EXISTS = True
except:
    #this will fail when using the module in wing
    pass


from .pigeon import *


class CascadeurPigeon(Pigeon):
    def __init__(self, *args, **kwargs):
        super(CascadeurPigeon, self).__init__(*args, **kwargs)
        self.known_pid = None


    @staticmethod
    def run_shell_command(cmd):
        #NOTE: don't use subprocess.check_output(cmd), because in python 3.6+ this error's with a 120 code.
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        stdout = stdout.decode()
        stderr = stderr.decode()

        print(stdout)
        print(stderr)
        if proc.returncode:
            raise Exception('Command Failed:\nreturn code:{0}\nstderr:\n{1}\n'.format(proc.returncode, stderr))

        return(stdout, stderr)


    @classmethod
    def get_temp_filename(cls):
        return('cascadeur_code.txt')
    
    
    def get_running_path(self):
        """Return the exe path of any running instance of cascadeur"""

        #we might already have a cached pid from wing.  let's try it first.
        if self.known_pid:
            try:
                process = psutil.Process(pid=self.known_pid)
                return process.exe()
            except:
                self.known_pid = None


        #let's search the running processes for cascadeur
        #ls: list = [] # since many processes can have same name it's better to make list of them
        for p in psutil.process_iter(['name', 'pid']):
            if p.info['name'] == 'cascadeur.exe':
                #we can only have one running process of cascadeur
                return psutil.Process(p.info['pid']).exe()

        return ''


    def can_dispatch(self):
        """Check if conditions are right to send code to application
        
        can_dispatch() is used to determine what dispatcher wing will use
        with when there's no active dispatcher found.
        """
        exe_path = self.get_running_path()
        return len(exe_path) > 0
    
    
    def owns_process(self, process):
        """Returns true if the process is the pigeons target application
        
        This is used when an external application is connects to wing. When True
        is returned the Dispatcher becomes the active dispatcher to send commands
        to.
        
        Args:
            process (psutils.Process) : The node to remove the data from.  
        """
        valid_process = 'cascadeur' in process.name()
        
        if valid_process:
            self.known_pid = process.pid
            
        return valid_process
    

    @classmethod
    def post_module_import(cls, module):
        """call the run function on the imported module if it exists
        
        We'll assume that if the run() takes any arguments that the first
        argument is the current scene, since this is the cascadeur standard.
        """

        print("Calling post module import")
        if hasattr(module, 'run'):
            signature = inspect.signature(module.run)
            if signature.parameters:
                if CSC_EXISTS:
                    scene = csc.app.get_application().get_scene_manager().current_scene()
                    module.run(scene)
                else:
                    module.run(None)
            else:
                module.run()


    def send(self, highlighted_text, module_path, file_path, doc_type):            
        if highlighted_text:
            command_string = highlighted_text
        else:
            command_string = u"import wingcarrier.pigeons; wingcarrier.pigeons.CascadeurPigeon.receive(\'{}\',\'{}\')".format(module_path, file_path)
            
        self.send_python_command(command_string)  
          
    
    def send_python_command(self, command_string):
        exe_path = self.get_running_path()
        if not exe_path:
            print('No instance of cascadeur is running')
            return
        
        command = '{}&--run-python-code&{}'.format(exe_path, command_string)
        CascadeurPigeon.run_shell_command(command.split('&'))        


    @staticmethod
    def receive(module_path, file_path):
        if not module_path:
            CascadeurPigeon.read_file(file_path)
        else:
            CascadeurPigeon.import_module(module_path, file_path)


import site
import sys
import subprocess
import socket
import os
import tempfile
from pathlib import Path

import wingapi

#we need to dynamically add our dispatchers package
_file_path = os.path.dirname(__file__)
_file_path = os.path.dirname(_file_path)

if _file_path not in sys.path:
    sys.path.append(_file_path)
    
import pigeons
import pigeons.maya
import pigeons.cascadeur
sys.path.remove(_file_path)


PSUTILS_EXISTS = False
try:
    import psutil
    print("wing-carrier: psutil found")
    PSUTILS_EXISTS = True
except:
    _parent_dir = os.path.dirname(_file_path)
    _psutils_dir = os.path.exists(_parent_dir, 'psutil')
    if (os.path.exists(_psutils_dir)):
        sys.path.append(_parent_dir)
        print('wing-carrier: found psutil package at sibling location')
        import psutil
        PSUTILS_EXISTS = True
        sys.path.remove(_parent_dir)
    else:
        print('wing-carrier: psutil missing')
        PSUTILS_EXISTS = False
        


CARRIERS = [
    pigeons.maya.MayaPigeon(),
    pigeons.cascadeur.CascadeurPigeon()
]
"""The global list of all carrier pigeons that can be dispatched

If a new Pigeon sub-class is created, it must also be added to this list to
be used.
"""

_CLASS_INSTANCE_MAPPING = {item.__class__.__name__: item for item in CARRIERS}
_ACTIVE_CARRIER: pigeons.pigeon.Pigeon = None
_DEBUG_CARRIER: pigeons.pigeon.Pigeon = None


def _get_document_text():
    """Based on the Wing API returns (selected text, doctype) """
    
    editor = wingapi.gApplication.GetActiveEditor()
    if editor is None:
        return ''
    
    doc = editor.GetDocument()
    doc_type = doc.GetMimeType()
    start, end = editor.GetSelection()
    txt = doc.GetCharRange(start, end)
    return (txt, doc_type)



def _get_module_info():
    """Returns the module namespace and the path to the file."""
    
    def _add_parent_module(name, path):
        found = False
        parent_module = os.path.join(path.parent, '__init__.py')
    
        if os.path.exists(parent_module):
            path = path.parent
            name = os.path.basename(path) + "." + name
            found = True
    
        return (found, name, path)    
    

    editor = wingapi.gApplication.GetActiveEditor()
    if editor is None:
        return ''

    doc = editor.GetDocument()
    full_path = doc.GetFilename()

    name = os.path.basename(full_path).split('.')[0]
    path = Path(full_path)
    loop = True
    count = 0
    while loop and count < 20:
        count += 1
        loop, name, path = _add_parent_module(name, path)

    #special case for when someone executes an __init__ file
    if (name.endswith(".__init__")):
        name = name.removesuffix(".__init__")

    return (name, full_path)



def _find_process_owner(process):
    global CARRIERS
    
    for dis in CARRIERS:
        if dis.owns_process(process):  
            return dis
            
    return None  
    

    #if _ACTIVE_DISPATCHER is None or not _ACTIVE_DISPATCHER.owns_process(process):
        #_ACTIVE_DISPATCHER = None
        
        #for dis in CARRIERS:
            #if dis.owns_process(process):  
                #_ACTIVE_DISPATCHER = dis
                #break
            
    ##we'll always the last active dispatcher at the top of our dispatcher list
    #if _ACTIVE_DISPATCHER and CARRIERS[0] !=  _ACTIVE_DISPATCHER:
        #idx = CARRIERS.index(_ACTIVE_DISPATCHER)
        #CARRIERS.pop(idx)
        #CARRIERS.insert(0, _ACTIVE_DISPATCHER)
                


def _find_best_process():
    valid_dispatchers = []
    
    for dis in CARRIERS:
        if dis.can_dispatch():
            valid_dispatchers.append(dis)
            
    if not valid_dispatchers:
        return None
    
    if len(valid_dispatchers) > 1:
        print("Mutliple dispatchers found. Using first one :{}".format(valid_dispatchers))
        
    return valid_dispatchers[0]



#def _activate_by_process(process):
    #"""If we're actively debugging try to find the process owner"""
    #global _ACTIVE_CARRIER

    #if process:
        #_find_process_owner(process)
    #else:
        #_ACTIVE_CARRIER = None
        
        

def dispatch_carrier(carrier: pigeons.pigeon.Pigeon = None):
    """Used to send the data to an external app based on a set of rules.
    
    When no carrier is provided the target carrier will be the last carrier
    dispatched. To provide a carrier use disptach methods such as
    dispatch_maya() or dispatch_cascadeur().
    
    When a debug connection is established that app's pigeon
    will take priority while connected (if no carrier is provided) otherwise
    the last active carrier will be used.
    
    If no carrier can be determined (or the previous carrier is no longer
    valid) then the function will attempt to find the best process to
    communicate with via Pigeon.can_dispatch().
    
    args:
        carrier (Pigeon)(Optional) : a specific pigeon to become the active carrier
    """
    global CARRIERS, _ACTIVE_CARRIER, _DEBUG_CARRIER
    
    target_carrier = None
    if carrier is None:
        if _DEBUG_CARRIER:
            target_carrier = _DEBUG_CARRIER
        else:
            target_carrier = _ACTIVE_CARRIER
    else:
        _ACTIVE_CARRIER = carrier
        target_carrier = _ACTIVE_CARRIER
     
    #A previously valid carrier now might not be valid, so it's
    #ensure our target is good and replace it if not.
    if not target_carrier or not target_carrier.can_dispatch():
        _ACTIVE_CARRIER = _find_best_process()
        target_carrier = _ACTIVE_CARRIER
        
    #We'll always move the last valid carrier to the top of the list
    #so it have priority when searching for a new carrier.
    if target_carrier and CARRIERS[0] != target_carrier:
        idx = CARRIERS.index(target_carrier)
        CARRIERS.pop(idx)
        CARRIERS.insert(0, target_carrier)
        
        
    if target_carrier is not None:
        highlighted_text, doc_type = _get_document_text()
        module_path, file_path = _get_module_info()
        file_path = file_path.replace("\\", "/")
        print('module path:{} full path:{}'.format(module_path, file_path))        
        
        _ACTIVE_CARRIER.send(highlighted_text, module_path, file_path, doc_type)
    else:
        print("No application to dispatch to!")
        
        

def dispatch_maya():
    dispatch_carrier(carrier=_CLASS_INSTANCE_MAPPING['MayaPigeon'])


def dispatch_cascadeur():
    dispatch_carrier(carrier=_CLASS_INSTANCE_MAPPING['CascadeurPigeon'])
        


#-----------WIN-IDE signal slots for active debug is below this line--------------
      
def _get_debug_process(current_run_state=None) :
    if current_run_state is None:
        debugger = wingapi.gApplication.GetDebugger()
        current_run_state = debugger.GetCurrentRunState()

    if not current_run_state:
        print("no debug run state found")
        return None
    
    pid = current_run_state.GetProcessID()
    process = None
    
    if PSUTILS_EXISTS:
        process: psutil.Process = psutil.Process(pid=pid)
        print('connected to PID:{}   name:{}   exe:{}'.format(process.pid, process.name(), process.exe()))
    else:
        print('wing-carrier: psutils missing')

    return process


        
def _debugger_connected(*args, **kwargs):
    global _DEBUG_CARRIER
    if args:
        process = _get_debug_process(args[0])
        if process is not None:
            _DEBUG_CARRIER = _find_process_owner(process)
        else:
            _DEBUG_CARRIER = None


    
def _debugger_changed(*args, **kwargs):
    global _DEBUG_CARRIER
    if not args:
        _DEBUG_CARRIER = None



debugger = wingapi.gApplication.GetDebugger()
debugger.Connect('new-runstate', _debugger_connected)
debugger.Connect('current-runstate-changed', _debugger_changed)
